// data segment descriptor
class SegmentData {
    int seq_num;
    byte [] data;
    int length;

    SegmentData(int maxlen) {
	data = new byte[maxlen];
    }
}

// Utility class to schedule timeouts within a TCP Sender.  This
// timeout scheduler maintains a single timeout. The last scheduled
// timeout is the one that gets executed by the sender.  A
// cancellation cancels all timeouts (because there is only one).
// 
class TimeoutScheduler implements Event {
    TimeoutScheduler(TCPSender s) {
	sender = s;
    }

    public void process() {
	if (scheduled) {
	    if (time <= Simulation.time()) {
		scheduled = false;
		sender.process_timeout();
	    }
	    else
		Simulation.schedule_event(time - Simulation.time(), this);
	}
    }

    public void schedule(double t) {
	if (scheduled && Simulation.time() + t > time)
	    return;
	time = Simulation.time() + t;
	scheduled = true;
	Simulation.schedule_event(t, this);
    }

    public void cancel() {
	scheduled = false;
    }

    private TCPSender sender;
    private boolean scheduled;
    private double time;
}

public class TCPSender implements  TransportSender, TransportReceiver {
    ApplicationSender app;
    Link link;

    int cnwd;			// congestion window
    double timeout = 1.0;	// timeout interval

    int src_port;
    int dst_port;
    int seq_num;		// next available sequence number
    int base;			// first pending (unacknowledged) sequence number
    int w_size;			// total amount of pending data

    int state;
    static final int CLOSED = 0;
    static final int SYN_SENT = 1;
    static final int ESTABLISHED = 2;
    static final int FIN_WAIT_1 = 3; // not yet implemented
    static final int FIN_WAIT_2 = 4; // not yet implemented
    static final int TIME_WAIT = 5;  // not yet implemented

    // list of pending data segments
    private java.util.LinkedList<SegmentData> data_list;

    private TimeoutScheduler timeout_scheduler;

    TCPSender(int src, int dst) {
	cnwd = 10000;		// default window size
	link = null;
	src_port = src;
	dst_port = dst;
	state = CLOSED;
	seq_num = 0;
	base = 0;
	w_size = 0;
	data_list = new java.util.LinkedList<SegmentData>();
	free_list = new java.util.LinkedList<SegmentData>();
	timeout_scheduler = new TimeoutScheduler(this);
    }

    // we maintain a list of free buffers ready to be recycled
    private java.util.LinkedList<SegmentData> free_list;

    private void recycle_buffer(SegmentData b) {
	free_list.add(b);
    }

    private SegmentData get_new_buffer() {
	if (free_list.isEmpty())
	    return new SegmentData(link.mtu() - 20); // TCP header length = 20 bytes
	else 
	    return free_list.remove();
    }

    void connect(Link l) {
	link = l;
    }

    void connect(ApplicationSender a) {
	app = a;
    }

    public void process_timeout() {	// action for "timeout" event on this sender
	System.out.printf("%f ", Simulation.time());
	System.out.println("TCPSender: timeout");

	switch (state) {
	case ESTABLISHED:
	    if (!data_list.isEmpty())
		timeout_scheduler.schedule(timeout);
	    for (SegmentData buf : data_list) {
		TCPPacket p = new TCPPacket(buf.data, buf.length);
		p.seq_num = buf.seq_num;
		p.src_port = src_port;	// sender port
		p.dst_port = dst_port;	// receiver port
		System.out.printf("%f ", Simulation.time());
		System.out.println("TCPSender: resending packet seq_num=" + p.seq_num);
		link.send(p);
	    }
	    break;
	case SYN_SENT:
	    TCPPacket p = new TCPPacket(null, 0);
	    timeout_scheduler.schedule(timeout);
	    p.syn = true;
	    p.seq_num = seq_num;
	    p.src_port = src_port;
	    p.dst_port = dst_port;
	    link.send(p);
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("TCPSender: resent SYN packet seq_num=" + p.seq_num);
	    break;
	}
    }

    public boolean send(byte [] data, int size) {
	if (state != ESTABLISHED)
	    return false;

	if (w_size > cnwd)
	    return false;

	int offset = 0;
	while (offset < size) {
	    // we might send multiple segments for this application request
	    SegmentData buf = get_new_buffer();
	    int i;
	    for (i = 0; i < buf.data.length && offset < size; ++i, ++offset)
		buf.data[i] = data[offset];
	    buf.length = i;
	    buf.seq_num = seq_num;
	    if (data_list.isEmpty())
		timeout_scheduler.schedule(timeout);

	    data_list.addLast(buf);

	    TCPPacket p = new TCPPacket(buf.data, i);
	    p.seq_num = seq_num;
	    p.src_port = src_port;
	    p.dst_port = dst_port;
	    link.send(p);
	    w_size += i;
	    System.out.printf("%f ", Simulation.time());
	    System.out.println("TCPSender: send data packet seq_num=" + seq_num + " w_size=" + w_size + "/" + cnwd); 
	    seq_num = TCP.next_seq(seq_num, i);
	}
	// we only block the sender application only *after* we exceed
	// the congestion window size.
	if (w_size > cnwd)
	    app.stop_sender();

	return true;
    }

    public void deliver(Packet packet) {
	if (packet instanceof TCPPacket) {
	    TCPPacket p = (TCPPacket)packet;
	    switch (state) {
	    case SYN_SENT:
		if (p.syn && p.ack && p.ack_num == TCP.next_seq(seq_num, 1)) {
		    System.out.printf("%f ", Simulation.time());
		    System.out.println("TCPSender: got SYN,ACK packet ack_num=" + p.ack_num); 
		    seq_num = TCP.next_seq(seq_num, 1);
		    base = seq_num;
		    p.syn = false;
		    p.ack = true;
		    p.ack_num = TCP.next_seq(p.seq_num, 1);
		    p.seq_num = seq_num;
		    p.app_data = null;
		    p.app_size = 0;
		    p.src_port = src_port;
		    p.dst_port = dst_port;
		    link.send(p);
		    timeout_scheduler.cancel(); // <-- set in case the SYN gets dropped
		    state = ESTABLISHED;
		    app.continue_sender();
		}
		break;
	    case ESTABLISHED:
		if (p.ack) {
		    if (seq_num >= base) {
			if (p.ack_num > seq_num || p.ack_num < base) {
			    System.out.printf("%f ", Simulation.time());
			    System.out.println("TCPSender: spurious ACK ack_num=" + p.ack_num); 
			    // spurious ACK
			    return;
			}
			if (p.ack_num == base) {
			    // duplicate ACK, ignore for now
			    System.out.printf("%f ", Simulation.time());
			    System.out.println("TCPSender: duplicate ACK base=" + base + " ack_num=" + p.ack_num);
			    return;
			}
		    } else {
			if (p.ack_num > seq_num && p.ack_num < base) {
			    System.out.printf("%f ", Simulation.time());
			    System.out.println("TCPSender: spurious ACK ack_num=" + p.ack_num); 
			    // spurious ACK
			    return;
			}
			if (p.ack_num == base) {
			    // duplicate ACK, ignore for now
			    System.out.printf("%f ", Simulation.time());
			    System.out.println("TCPSender: duplicate ACK base=" + base + " ack_num=" + p.ack_num);
			    return;
			}
		    }
		    System.out.printf("%f ", Simulation.time());
		    System.out.println("TCPSender: good ACK base=" + base + " ack_num=" + p.ack_num); 
		    while (!data_list.isEmpty()
			   && TCP.seq_diff(p.ack_num, base) >= data_list.getFirst().length) {
			System.out.printf("%f ", Simulation.time());
			System.out.println("TCPSender: done with base=" + base);
			SegmentData b = data_list.removeFirst();
			base = TCP.next_seq(base, b.length);
			w_size -= b.length;
			recycle_buffer(b);
		    }
		    if (data_list.isEmpty())
			timeout_scheduler.cancel();
		    else
			timeout_scheduler.schedule(timeout);
		    if (w_size < cnwd)
			app.continue_sender();
		}
		break;
	    }
	}
    }

    public void open() {
	// Send SYN
	app.stop_sender();
	seq_num = new java.util.Random().nextInt();
	TCPPacket p = new TCPPacket(null, 0);
	timeout_scheduler.schedule(timeout);
	p.syn = true;
	p.seq_num = seq_num;
	p.src_port = src_port;
	p.dst_port = dst_port;
	link.send(p);
	state = SYN_SENT;
	System.out.printf("%f ", Simulation.time());
	System.out.println("TCPSender: sent SYN packet seq_num=" + p.seq_num); 
    }

    public void close() { 	// not yet implemented
    }
}
